/* mkfig1.c - compute BM impedance for MOH83 Fig. 1 */

#include <stdio.h>
#include <math.h>

typedef struct {double x, y;} complex;

int m = 8;
int n = 409;
double xl = 2.55;
double yw = 0.1;
double zh = 0.1;

/* set of complex arithmetic functions */

complex
c_add(complex z1, complex z2)		/* complex add */
{
    complex z3;

    z3.x = z1.x + z2.x;
    z3.y = z1.y + z2.y;

    return (z3);
}

complex
c_sub(complex z1, complex z2)		/* complex subtract */
{
    complex z3;

    z3.x = z1.x - z2.x;
    z3.y = z1.y - z2.y;

    return (z3);
}

complex
c_mul(complex z1, complex z2)		/* complex multiply */
{
    complex z3;

    z3.x = z1.x * z2.x - z1.y * z2.y;
    z3.y = z1.y * z2.x + z1.x * z2.y;

    return (z3);
}

complex
c_div(complex z1, complex z2)		/* complex divide */
{
    complex z3;
    double ms;

    ms = z2.x * z2.x + z2.y * z2.y;
    z3.x = (z1.x * z2.x + z1.y * z2.y) / ms;
    z3.y = (z1.y * z2.x - z1.x * z2.y) / ms;

    return (z3);
}

complex
c_rcp(complex z)			/* complex reciprocal */
{
    double ms;

    ms = z.x * z.x + z.y * z.y;
    z.x = z.x / ms;
    z.y = -z.y / ms;

    return (z);
}

complex
c_ri(double zr, double zi)		/* complex (real,imaginary) */
{
    complex z;

    z.x = zr;
    z.y = zi;

    return (z);
}

/* c_krm - compute impedance z(f) = k/(iw) + r + m*(iw)
 *          where w=2*pi*f
 */
complex
c_krm(double k, double r, double m, double f)
{
    complex z;
    double w;
    static double tpi = 2 * 3.1415927;

    w = tpi * f;
    z.x = r;
    z.y = m * w - k / w;

    return (z);
}

/* c_krmx - compute impedance z(f,x) = k(x)/(iw) + r(x) + m(x)*(iw) 
 *          where w=2*pi*f
 *                k(x)=k0*exp(ke*x)
 *                r(x)=r0*exp(re*x)
 *                m(x)=m0*exp(me*x)
 */
complex
c_krmx(double ko, double ke, double ro, double re, double rc, double rd,
    double mo, double me, double f, double x)
{
    complex z;
    double w;
    static double tpi = 2 * 3.1415927;

    w = tpi * f;
    z.x = ro * exp (re * x) + rc * exp (rd * x);
    z.y = mo * exp (me * x) * w - ko * exp (ke * x) / w;

    return (z);
}

/* imped - compute partition impedance */

complex
imped(double f, double x)
{
    complex z1, z2, z3, za, zb;

    static double k1o = 1e9, k1e = -2.4;    /* stiffness component of Z1 */
    static double r1o = 400, r1e = 0;       /* damping component of Z1 */
    static double r1c = 1e3, r1d = -1.2;    /* r1=r1o*e^r1e*x + r1c*e^r1d*x */
    static double m1o = 0.001, m1e = 1.2;   /* mass component of Z1 */
    static double k2o = 250, k2e = -2.2;    /* stiffness component of Z2 */
    static double r2o = 1e-4, r2e = 1.1;    /* damping component of Z2 */
    static double r2c = 8e-4, r2d = -2.2;   /* r2=r2o*e^r2e*x + r2c*e^r2d*x */
    static double m2o = 3e-9, m2e = 2.2;    /* mass component of Z2 */
    static double k3o = 0, k3e = 0;         /* stiffness component of Z3 */
    static double r3o = 1, r3e = 0;         /* damping component of Z3 */

    /* Cochlear partition */

    z1 = c_krmx(k1o, k1e, r1o, r1e, r1c, r1d, m1o, m1e, f, x);
    z2 = c_krmx(k2o, k2e, r2o, r2e, r2c, r2d, m2o, m2e, f, x);
    z3 = c_krmx(k3o, k3e, r3o, r3e, 0, 0, 0, 0, f, x);
    /* za = z3 * z3 / z2 */
    za = c_mul(z3, c_div(z3, z2));
    /* zb = z1 - za */
    zb = c_sub(z1, za);

    return (zb);
}

/* dbph - compute magnitude (in dB) and phase (unwrapped) */

complex
dbph(complex p)
{
    double mg, db, ph, lp = 0;
    static double tpi = 2 * 3.1415927;
    
    mg = hypot(p.x, p.y);
    if (mg > 1e-30) {
        db = 20 * log10(mg);
        ph = atan2(p.y, p.x) / tpi;
    } else {
        db = -600;
        ph = 0;
    }
    if (ph > lp) {
        ph -= floor(ph - lp + 0.5);
    } else {
        ph += floor(lp - ph + 0.5);
    }
    lp = ph;
    p.x = db;
    p.y = ph;

    return (p);
}

/* wrdata - write data file header*/

void
wrhead(FILE *fp, char *fn)
{
    fprintf(fp, "; %s - BM impedance for MOH83\n", fn);
    fprintf(fp, "; m=%d n=%d xl=%.2f yw=%.2f zh=%.2f\n", m, n, xl, yw, zh);
    fprintf(fp, ";         ");
    fprintf(fp, "|                  Zbm(0.2)                 ");
    fprintf(fp, "|                  Zbm(0.4)                 ");
    fprintf(fp, "|                  Zbm(0.6)                 ");
    fprintf(fp, "|                  Zbm(0.8)                 ");
    fprintf(fp, "|\n");
    fprintf(fp, ";    freq ");
    fprintf(fp, "|       [RE]         [IM]     [dB]     [ph] ");
    fprintf(fp, "|       [RE]         [IM]     [dB]     [ph] ");
    fprintf(fp, "|       [RE]         [IM]     [dB]     [ph] ");
    fprintf(fp, "|       [RE]         [IM]     [dB]     [ph] ");
    fprintf(fp, "|\n");
}

/* wrdata - write BM imdedance data */

void
wrdata(FILE *fp, double f)
{
    complex z;
    double x;
    int i;

    fprintf(fp, " %8.4f", f / 1000);

    for (i = 0; i < 4; i++) {

	x = 0.2 * xl * (i + 1);

        /* BM impedance */
        z = imped(f, x);
        fprintf(fp, " %12.4g %12.4g", z.x, z.y);

        /* convert to magnitude and phase */
        z = dbph(z);
        fprintf(fp, " %8.1f %8.3f", z.x, z.y);
    }

    fprintf(fp, "\n");

}


int
main()
{
    double f, df=48.8;
    int i;
    FILE *fp;
    static char fn[80] = "fig1.dat";

    fp = fopen("fig1.dat", "wt");
    wrhead(fp, fn);
    for (i = 1; i < 600; i++) {
	f = i * df;
        wrdata(fp, f);
    }
    fclose(fp);

    return (0);
}
