/* tdmini.c - initializing subroutines for time-domain model */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "tdmdef.h"
#include "cdata.h"
#include "cimped.h"
#include "cmider.h"
#include "cparam.h"
#include "ctdmio.h"
#include "cstate.h"

#define limit(a,b,c)	(((a)>(b))?(a):((c)<(b))?(c):(b))

/* xinterp - inperpolated logarithmically */

double
xinterp(double x, float *fp)
{
    double al, bl, cl, xl, pv;

    if (fp[0] == 0) {
	pv = 0;
    } else if (fp[1] == 0 && fp[2] == 0) {
	xl = pa_.xl * x;
	pv = fp[0] * exp(fp[4] * xl);
    } else {
        al = (fp[0] > 0) ? log(fp[0]) : -99;
        bl = (fp[1] > 0) ? log(fp[1]) : -99;
        cl = (fp[2] > 0) ? log(fp[2]) : -99;
        xl = (x < 0.5) ? (2*(bl-al)*x+al) : (2*(cl-bl)*x+(2*bl-cl));
        pv = exp(xl);
    }

    return (pv);
}

/* midear - initialize middle-ear model */

static void
midear()
{
    float   cep, rep, a, b, c, d, det, ma, ra, rb, ka, kb, gme;
    float   adc, acd, aec, ace, acm, amx, amc, amd, ams;
    int     i, j;

    pa_.nmev = NME;
    me_ep = ME_EP;
    me_ed = ME_ED;
    me_ma = ME_MA;
    me_st = ME_ST; 
    ame[ME_EP] = pa_.adi;
    ame[ME_ED] = pa_.acp;
    ame[ME_MA] = pa_.ama;
    ame[ME_ST] = pa_.ast;

    cep = 0.1 * (pa_.bl / pa_.rvc);
    rep = 1e-9 * ((pa_.bl * pa_.bl) / pa_.rvc); 
    acd = pa_.acp / pa_.adi;
    ace = pa_.acp / pa_.aed;
    acm = pa_.acp / pa_.ama;
    adc = pa_.adi / pa_.acp;
    aec = pa_.aed / pa_.acp;
    amc = pa_.ama / pa_.acp;
    amd = pa_.ama / pa_.adi;
    amx = pa_.ama / pa_.aed;
    ams = pa_.ama / pa_.ast;
    gme = pa_.gmel;

    if (pa_.basacc && !pa_.couplr) {	/* stimulus to eardurm */
	ma = ra = rb = ka = kb = 0;
    } else {
	ma = pa_.mcp;
        ra = pa_.rcp + pa_.rfz;
        rb = pa_.rcp * adc;
        ka = pa_.kcp;
        kb = pa_.kcp * adc;
    }

    for (i = 0; i < NME; i++) {
        pedm[i] = pedr[i] = pedk[i] = smes[i] = 0;
        for (j = 0; j < NME; j++) {
	    smem[i][j] = smer[i][j] = smek[i][j] = 0;
	}
    } 

    smem[ME_EP][ME_EP] = pa_.mdi;
    smem[ME_ED][ME_ED] = ace * pa_.med + aec * ma;
    smem[ME_ED][ME_MA] = amx * pa_.med;
    smem[ME_MA][ME_ED] = amc * ma;
    smem[ME_MA][ME_MA] = pa_.mma;
    smem[ME_ST][ME_ST] = pa_.mst;

    smer[ME_EP][ME_EP] = pa_.rdi + rep + pa_.rcp * adc * adc;
    smer[ME_EP][ME_ED] = -pa_.rcp * adc;
    smer[ME_ED][ME_EP] = -aec * rb;
    smer[ME_ED][ME_ED] = ace * pa_.red + aec * ra;
    smer[ME_ED][ME_MA] = -amx * pa_.red;
    smer[ME_MA][ME_EP] = -amc * rb;
    smer[ME_MA][ME_ED] = amc * ra;
    smer[ME_MA][ME_MA] = pa_.rma +  gme * gme * pa_.rim;
    smer[ME_MA][ME_ST] = -gme * pa_.rim;
    smer[ME_ST][ME_MA] = -gme * pa_.rim;
    smer[ME_ST][ME_ST] = pa_.rst + pa_.rim;

    smek[ME_EP][ME_EP] = pa_.kdi + adc * adc * pa_.kcp;
    smek[ME_EP][ME_ED] = -adc * pa_.kcp;
    smek[ME_ED][ME_EP] = -aec * kb;
    smek[ME_ED][ME_ED] = ace * pa_.ked + aec * ka;
    smek[ME_ED][ME_MA] = -amx * pa_.ked;
    smek[ME_MA][ME_EP] = -amc * kb;
    smek[ME_MA][ME_ED] = amc * ka;
    smek[ME_MA][ME_MA] = pa_.kma +  gme * gme * pa_.kim;
    smek[ME_MA][ME_ST] = -(gme * pa_.kim);
    smek[ME_ST][ME_MA] = -(gme * pa_.kim);
    smek[ME_ST][ME_ST] = pa_.kst + pa_.kim;

    /* divide sme by ame */

    for (i = 0; i < NME; i++) {
        for (j = 0; j < NME; j++) {
    	    smem[i][j] /= ame[i];
    	    smer[i][j] /= ame[i];
    	    smek[i][j] /= ame[i];
	}
    }

    /* invert mass matric (smem) */

    a = smem[ME_ED][ME_ED];
    b = smem[ME_ED][ME_MA];
    c = smem[ME_MA][ME_ED];
    d = smem[ME_MA][ME_MA];
    det = a * d - b * c;
    if (det < 1e-9) {
        fprintf(stderr, "bad values for middle-ear parameters: det=%g\n", det);
	exit(1);
    }
    smem[ME_EP][ME_EP] = 1  / smem[ME_EP][ME_EP];
    smem[ME_ED][ME_ED] = d / det;
    smem[ME_ED][ME_MA] = -b / det;
    smem[ME_MA][ME_ED] = -c / det;
    smem[ME_MA][ME_MA] = a / det;
    smem[ME_ST][ME_ST] = 1  / smem[ME_ST][ME_ST];

    /* apply stimulus */

    if (pa_.basacc) {
	if (pa_.couplr) {		/* to earphone */
            smes[ME_EP] = cep / pa_.adi;
	} else {			/* to eardrum */
            smes[ME_ED] = 1;
            smes[ME_MA] = 1;
	}
    }
    if (pa_.basacc && !pa_.couplr) {
        peds = 1;
    } else {
        peds = 0;
        pedm[ME_ED] = (pa_.med / pa_.aed) * (pa_.acp / pa_.aed);
        pedm[ME_MA] = -(pa_.med / pa_.aed) * (pa_.ama / pa_.aed);
        pedr[ME_ED] = (pa_.red / pa_.aed) * (pa_.acp / pa_.aed);
        pedr[ME_MA] = -(pa_.red / pa_.aed) * (pa_.ama / pa_.aed);
        pedk[ME_ED] = (pa_.ked / pa_.aed) * (pa_.acp / pa_.aed);
        pedk[ME_MA] = -(pa_.ked / pa_.aed) * (pa_.ama / pa_.aed);
    }
}
/* mechcp - compute mechanical data */

void
mechcp()
{
    int     i, i1, i2, ii, nn;
    float   xol, oonn, apr, app, apt;
    static float apq = 1.453, apf = 1000, apv = 3.463e-6;
    static float tpi = 6.28318530718;
    static float eps = 1e-9;

    /* force n into range */

    pa_.m = limit(1, pa_.m, Mmax);
    pa_.n = limit(1, pa_.n, Nmax);
    fprintf(lfp, "  m,n,xl,yw,zh     =%6d%6d%8.2f%8.2f%8.2f\n",
	pa_.m, pa_.n, pa_.xl, pa_.yw, pa_.zh);
    nn = pa_.n - 1;
    oonn = (nn > 0) ? 1.0 / nn : 0;

    /* force logical variables to be consistent */
    if (pa_.midear == FALSE) {
	pa_.couplr = FALSE;
        pa_.basacc = TRUE;
    }

    /* useful constructs describing cochlear dimensions */

    dx = (pa_.n > 1) ? pa_.xl / (pa_.n - 1) : pa_.xl;
    dz = (pa_.m > 1) ? pa_.zh / (pa_.m - 1) : pa_.zh;
    if (pa_.mmeq == 0)
        nltype = (pa_.g1a == 0 && pa_.g2a == 0) ? 0 : 1;
    else
        nltype = (pa_.d3max <= 0) ? 0 : 1;

    /* compute partition */

    if (pa_.m1a < eps) {
	pa_.m1a = eps;
	pa_.m1b = pa_.m1c = pa_.m1e = 0;
    }
    if (pa_.m2a < eps && pa_.mmeq != 0) {
	pa_.m2a = eps;
	pa_.m2b = pa_.m2c = pa_.m2e = 0;
    }

    for (i = 0; i <= nn; i++) {
	xol = i * oonn;
	z1k[i] = xinterp(xol, &pa_.k1a);	/* k1 */
	z1m[i] = xinterp(xol, &pa_.m1a);	/* m1 */
	z1r[i] = xinterp(xol, &pa_.r1a);	/* r1 */
	z2k[i] = xinterp(xol, &pa_.k2a);	/* k2 */
	z2m[i] = xinterp(xol, &pa_.m2a);	/* m2 */
	z2r[i] = xinterp(xol, &pa_.r2a);	/* r2 */
	z3k[i] = xinterp(xol, &pa_.k3a);	/* k3 */
	z3r[i] = xinterp(xol, &pa_.r3a);	/* r3 */
	z4k[i] = xinterp(xol, &pa_.k4a);	/* k4 */
	z4r[i] = xinterp(xol, &pa_.r4a);	/* r4 */
	z5k[i] = xinterp(xol, &pa_.k5a);	/* k5 */
	g1[i] = xinterp(xol, &pa_.g1a);
	g2[i] = xinterp(xol, &pa_.g2a);
	t1[i] = xinterp(xol, &pa_.t1a);
    }

    /* BM roughness */

    if (pa_.k1r > 0) {
	double a, b;

	a = 1 - pa_.k1r;
	b = 2 * pa_.k1r / RAND_MAX;
        for (i = 0; i <= nn; i++) {
            z1k[i] *= a + b * random();
	}
    }

    /* helicotrema region */

    if (pa_.apxacc) {
        if (pa_.mhe < eps)
	    pa_.mhe = eps;
	ii = (int) (nn * (1 - pa_.xhe / pa_.xl) + 0.5); 
        for (i = ii; i <= nn; i++) {
            z1k[i] = pa_.khe;
            z1m[i] = pa_.mhe;
            z1r[i] = pa_.rhe;
            z2k[i] = z3k[i] = 0;
            z2m[i] = 1;
            z2r[i] = z3r[i] = 0;
	}
    }

    /* OHC lesion */

    if (pa_.xl2 > pa_.xl1 && pa_.xl1 >= 0) {
    	i1 = floor(pa_.xl1 / dx); 
    	i2 = ceil(pa_.xl2 / dx);
    	for (i = i1; i <= i2 && i < nn; i++)
    	    g1[i] = g2[i] = 0;
    }

    /* middle-ear and coupler data */

    midear();

    /* inner hair cell */

    if (pa_.tauhc == 0)
	pa_.tauhc = 0.5e-3;
    cc1 = exp(-pa_.dt / pa_.tauhc);
    cc2 = 1;
    //dvslo = (pa_.dhbmin >= pa_.dhbmax) ? 1 :
	//(pa_.vhcmax - pa_.vhcmin) / (pa_.dhbmax - pa_.dhbmin);
    //dvint = pa_.vhcmin - dvslo * pa_.dhbmin;
    app = tpi * apf * pa_.dt;
    apr = exp(-app * (apq - sqrt(apq * apq - 1)));
    apt = pa_.dt * pa_.ntsw;
    apk0 = (apt > 0) ? apv * apr * sin(app) / apt : 1;
    apk1 = 2 * apr * cos(app);
    apk2 = -(apr * apr);

}

/* state0 - set initial state */

void
state0()
{
    int     i;

    /* timing counters */
    pa_.itsc = 0;
    pa_.icyc = 0;
    pa_.itsf = 0;
    pa_.itsw = 0;

    /* middle ear and cochlea */
    for (i = 0; i < NSTV; i++) {
	st_.d_[i] = 0.0;
	st_.dd_[i] = 0.0;
    }
    st_.pedmax = 0;
    st_.pedsum = 0;
    st_.pedssq = 0;
    st_.pednum = 0;
    st_.wnu0 = 0;
    st_.wnu1 = 0;
    st_.wnap = 0;

    /* outer hair cell */

    for (i = 0; i < pa_.n; i++) {
	st_.vohc[i] = 0.0;
	st_.dvohc[i] = 0.0;
    }

    /* inner hair cell */
    for (i = 0; i < pa_.n; i++) {
	st_.cc[i] = 0.0;
	st_.qc[i] = 0.0;
	st_.sc[i] = 0.0;
    }

}
