/* tdmstm.c - stimulus subroutines for time-domain model */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <unistd.h>
#ifdef MSDOS
#include <sys/stat.h>
#else
#define O_BINARY    0
#endif
#include "tdmdef.h"
#include "cparam.h"
#include "ctdmio.h"

#define BS 	256		/* number of samples per disk block     */
#define SDF 	-32000		/* ILS sampled data file (magic number) */
#define MAXTSD  8192            /* maximum number of TSD time-steps     */
#define limit(a,b,c)	(((a)>(b))?(a):((c)<(b))?(c):(b))

int     opnblk(char *);
int     rdblk(short, short, short *, short);
void    clsblk(short);
void    error();
void    swab();

static double interp();
static double fetch();
static void compute_lpf();

static double rate = 0;
static double delt = 0;
static double ts = 1;
static double a1 = 0, a2 = 0, a3 = 0, a4 = 0;
static double p1 = 0, p2 = 0, p3 = 0, p4 = 0;
static double pi = 3.14159265359;
static long nsmp = 0;
static long nssc = 0;
static long sts = 0;
static short ifd = -1;
static short ipnt = 0;
static short iblk = 0;
static short jblk = 0;
static short ibuf[BS];
static short jbuf[BS];
static short swap_bytes = 0;
static short stim_type = 0;

static int n_up = 32;		/* upsample ratio */
static int n_fw = 8;		/* filter width */
static float *hh = NULL;	/* lpf impulse response */

static int
chk_ils()
{
    swap_bytes = 0;
    if (rdblk(ifd, 0, ibuf, 1) < 1)
	return (0);
    if (ibuf[62] == SDF)
	return (1);
    swap_bytes = 1;
    if (rdblk(ifd, 0, ibuf, 1) < 1)
	return (0);
    if (ibuf[62] == SDF)
	return (1);
    return (0);
}

static int
valid_mat_hdr()
{
    char    type[4];
    int     status = 0;
    long   *hdr = (long *) ibuf;

    if (hdr[0] < 0 || hdr[0] > 9999)
	return (0);
    type[0] = (hdr[0] / 1000) % 10;	/* M */
    type[1] = (hdr[0] / 100) % 10;	/* O */
    type[2] = (hdr[0] / 10) % 10;	/* P */
    type[3] = (hdr[0] / 1) % 10;	/* T */
    if (type[3] == 0)
	if (type[0] == 0 || type[0] == 1)
	    if (type[1] == 0 || type[1] == 1)
		if (type[2] >= 0 && type[2] <= 4)
		    if (hdr[3] == 0)
			if (hdr[1] > 0 && hdr[2] > 0)
			    if (hdr[1] == 1 || hdr[2] == 1)
			        if (hdr[4] > 0)
				    status = 1;
    return (status);
}

static int
chk_mat()
{
    swap_bytes = 0;
    if (rdblk(ifd, 0, ibuf, 1) < 1)
	return (0);
    if (valid_mat_hdr())
	return (1);
    swap_bytes = 1;
    if (rdblk(ifd, 0, ibuf, 1) < 1)
	return (0);
    if (valid_mat_hdr())
	return (1);
    return (0);
}
 
static int
chk_wav()
{
   long   *rif_hdr = (long *) ibuf;
   static long key[2] = {0x46464952, 0x45564157};

    swap_bytes = 0;
    if (rdblk(ifd, 0, ibuf, 1) < 1)
	return (0);
    if (rif_hdr[0] == key[0] && rif_hdr[2] == key[1])
	return (1);
    swap_bytes = 1;
    if (rdblk(ifd, 0, ibuf, 1) < 1)
	return (0);
    if (rif_hdr[0] == key[0] && rif_hdr[2] == key[1])
	return (1);

    return (0);
}

static int
chk_stm(char *fn, char *em)
{
    if (strcmp(fn, "im30") == 0) {
	stim_type = 4;
    } else if (access(fn, 0) != 0) {
	sprintf(em, "can't open stimulus file (%s)", fn);
        stim_type = 0;
    } else {
        if (ifd >= 0)
	    clsblk(ifd);
        if ((ifd = opnblk(fn)) < 0) {
	    sprintf(em, "reading stimulus header (%s)", fn);
            stim_type = 0;
	} else {
            if (chk_ils()) {
	        stim_type = 1;
            } else if (chk_mat()) {
	        stim_type = 2;
            } else if (chk_wav()) {
	        stim_type = 3;
            } else {
	        sprintf(em, "unknown stimulus format (%s)", fn);
                stim_type = 0;
            }
        }
    }
    return (stim_type);
}

int
stmfil(char *fn)
{
    char em[80];

    if (chk_stm(fn, em)) {
	strcpy(pa_.stimfn, fn);
	return (1);
    }
    return (0);
}

static void
ini_ils()
{
    short nblk, ipow, irat, majk;

    nblk = ibuf[5];
    ipow = ibuf[60];
    irat = ibuf[61];
    majk = ibuf[62];
    rate = irat * pow(10.0, (double) ipow);
    nsmp = nblk * ((long) BS);
}

#define MAXNCH 256

typedef struct {
    long data, next, rows, cols;
    short type, vers, bswp;
    char name[MAXNCH];
} MATHDR;

static long stm_doff = 0;
static int stm_dsiz = 0;
static int stm_dtyp = 0;

int
rd_mat_hdr(int fd, MATHDR *mh)
{
    int M, O, P, T, cmpx, nchr = 0;
    long hdr[12];
    static int dtyp[5] = {9, 8, 5, 3, 6};
    static int size[5] = {8, 4, 4, 2, 4};
    /* P= [0,1,2,3,4]=>[R*8,R*4,I*4,I*2,U*4] */

    if (read(fd, hdr, 20) < 20)
	return (0);
    M = (hdr[0] / 1000) % 10;
    O = (hdr[0] / 100) % 10;
    P = (hdr[0] / 10) % 10;
    T = (hdr[0] / 1) % 10;
    mh->rows = hdr[1];
    mh->cols = hdr[2];
    cmpx = hdr[3];
    nchr = hdr[4];
    if (nchr <= 0 || nchr > MAXNCH)
        return (0);
    if (read(fd, mh->name, nchr) < nchr)
        return (0);
    mh->data = lseek(fd, 0L, 1);
    mh->next = mh->data + mh->rows * mh->cols * size[P];
    mh->type = dtyp[P];

    return (1);
}

int
rd_stm_val(int fd, int p, double *v)
{
    unsigned char u1;
    double f8;
    float f4;
    long i4;
    short i2;
    unsigned long u4;
    unsigned short u2;

    switch (p) {
    case 2:	/* unsigned int_8 */
	if (read(fd, &u1, 1) < 1)
	    return (0);
	*v = u1;
	break;
    case 3:	/* signed int_16 */
	if (read(fd, &i2, 2) < 2)
	    return (0);
	*v = i2;
	break;
    case 4:	/* unsigned int_16 */
	if (read(fd, &u2, 2) < 2)
	    return (0);
	*v = u2;
	break;
    case 5:	/* signed int_32 */
	if (read(fd, &i4, 4) < 4)
	    return (0);
	*v = i4;
	break;
    case 6:	/* unsigned int_32 ??? */
	if (read(fd, &u4, 4) < 4)
	    return (0);
	*v = u4;
	break;
    case 8:	/* single float ??? */
	if (read(fd, &f4, 4) < 4)
	    return (0);
	*v = f4;
	break;
    case 9:	/* double float */
	if (read(fd, &f8, 8) < 8)
	    return (0);
	*v = f8;
	break;
    default:
	*v = 0;
    }

    return (1);
}

static void
ini_mat()
{
    MATHDR m;
    static int size[5] = {8, 4, 4, 2, 4};

    lseek(ifd, 0L, 0);	/* rewind file */
    rate = 20000;
    nsmp = 0;
    if (!rd_mat_hdr(ifd, &m))
        return;
    while (m.rows == 1 && m.cols == 1) {
        if (strcmp("rate", m.name) == 0) {
            lseek(ifd, m.data, 0);	/* skip to data value */
	    rd_stm_val(ifd, m.type, &rate);
	}
	lseek(ifd, m.next, 0);	/* skip to next variable */
	if (rd_mat_hdr(ifd, &m))
	    return;
    }
    nsmp = m.rows * m.cols;
    stm_doff = m.data;
    stm_dsiz = size[m.type];
    stm_dtyp = m.type;
}

static void
ini_wav()
{
    char   *s, *sbuf = (char *) ibuf;
    long    fmtsz, datsz, hdrsz, size, *lbuf = (long *) ibuf;
    struct {
        short frmt, chan, blk_algn, bit_samp;
        long samp_sec, byte_sec;
    } wav_hdr;

    lseek(ifd, 0L, 0);	/* rewind file */
    rate = 0;
    nsmp = 0;
    if (rdblk(ifd, 0, ibuf, 1) < 1)
	return;
    size = lbuf[1];
    fmtsz = lbuf[4];
    wav_hdr.frmt = ibuf[10];
    wav_hdr.chan = ibuf[11];
    wav_hdr.samp_sec = lbuf[6];    
    wav_hdr.byte_sec = lbuf[7];
    wav_hdr.blk_algn = ibuf[16];
    wav_hdr.bit_samp = ibuf[17];
    if (wav_hdr.bit_samp != 16) {
        printf("WAV file (%s) must have 16 bits/sample\n", pa_.stimfn);
        return;
    }
    hdrsz = 20 + fmtsz;
    datsz = 0;
    while (hdrsz < BS * 2) {
        s = sbuf + hdrsz;
        size = *((long *) (s + 4));
        hdrsz += 8;
        if (strncmp(s, "data", 4) == 0) {
            datsz = size;
            break;
        }
        hdrsz += size;
    }
    stm_doff = hdrsz;
    stm_dsiz = 2;
    stm_dtyp = 3;
    rate = wav_hdr.samp_sec;
    nsmp = datsz / stm_dsiz;
}

static void
ini_imp()
{
    rate = 1e5;
    nsmp = 2048;
}

static void
ini_stm()
{
    if (stim_type == 1)
	ini_ils();
    else if (stim_type == 2)
	ini_mat();
    else if (stim_type == 3)
	ini_wav();
    else if (stim_type == 4)
	ini_imp();
}

void
stmini()
{
    char em[80];

    if (chk_stm(pa_.stimfn, em) == 0)
	return;
    printf("simtulus file: %s\n", pa_.stimfn);
    ini_stm();
    delt = (rate > 0) ? 1 / rate : 1;
    if (nsmp > MAXTSD) {
	pa_.ntsf = MAXTSD;
	pa_.ntsw = (delt * nsmp) / (pa_.dt * MAXTSD) + 0.2;
    } else {
	pa_.ntsf = nsmp;
	pa_.ntsw = delt / pa_.dt + 0.2;
    }
    pa_.ntsc = pa_.ntsf * pa_.ntsw;
    pa_.ncyc = 1;
    pa_.ifrmt = 1;
    pa_.jfrmt = 6;
    nssc = pa_.ntsc * ts + 0.2;
    pa_.f1 = pa_.f2 = pa_.f3 = pa_.f4 = 0;
    pa_.l1 = pa_.l2 = pa_.l3 = pa_.l4 = 0;
}

void
stmopn()
{
    char em[80];
    int i1, i2, i3, i4;
    long ns;
    double tc;
    
    if (pa_.f1 > 0 || pa_.f2 > 0 || pa_.f3 > 0 || pa_.f4 > 0) {
        ns = limit(1, pa_.ntsf * pa_.ntsw, 1000000);
        tc = pa_.dt * ns;
        i1 = (int) (pa_.f1 * tc + 0.5);
        i2 = (int) (pa_.f2 * tc + 0.5);
        i3 = (int) (pa_.f3 * tc + 0.5);
        i4 = (int) (pa_.f4 * tc + 0.5);
        p1 = (2 * pi * i1) / ns;
        p2 = (2 * pi * i2) / ns;
        p3 = (2 * pi * i3) / ns;
        p4 = (2 * pi * i4) / ns;
        a1 = pow(10.0, pa_.l1 / 20);
        a2 = pow(10.0, pa_.l2 / 20);
        a3 = pow(10.0, pa_.l3 / 20);
        a4 = pow(10.0, pa_.l4 / 20);
        pa_.f1 = tc > 0 ? i1 / tc : 0;
        pa_.f2 = tc > 0 ? i2 / tc : 0;
        pa_.f3 = tc > 0 ? i3 / tc : 0;
        pa_.f4 = tc > 0 ? i4 / tc : 0;
        pa_.ntsc = ns;
        fprintf(lfp, "  f1,f2,f3,f4         = %8.0f%8.0f%8.0f%8.0f\n",
	    pa_.f1, pa_.f2, pa_.f3, pa_.f4);
        fprintf(lfp, "  l1,l2,l3,l4         = %8.0f%8.0f%8.0f%8.0f\n",
	    pa_.l1, pa_.l2, pa_.l4, pa_.l4);
        return;
    }
    if (pa_.stimfn[0] == '\0') {
	ifd = -1;
	return;
    }
    if (chk_stm(pa_.stimfn, em) == 0) {
	error(em);
    }
    ini_stm();
    delt = (rate > 0) ? 1 / rate : 1;
    ts = pa_.dt / delt;
    sts = nsmp;
    nssc = pa_.ntsc * ts + 0.2;
    fprintf(lfp,
	"  stimfn=%s,  stgain=%.3e,  delt=%.3e,  nsmp=%ld, nssc=%ld\n",
	pa_.stimfn, pa_.stgain, delt, nsmp, nssc);
    ipnt = 0;
    iblk = 0;
    jblk = 0;

    if (n_up > 0) {
	ts *= n_up;
	sts *= n_up;
	compute_lpf();
    }
}

void
stmcls()
{
    if (ifd >= 0) {
	clsblk(ifd);
	ifd = -1;
    }
}

/* stimul - return stimulus value for model time step 'mts' */

double
stimul(mts)
long    mts;
{
    long    k;
    double  v, g, t;
    static double v0 = 1e4;

    if (p1 > 0 || p2 > 0 || p3 > 0 || p4 > 0) {
    	k = mts % pa_.ntsc;
	v = 0;
        if (p1 > 0)
    	    v += a1 * sin(p1 * k);
        if (p2 > 0)
            v += a2 * sin(p2 * k);
        if (p3 > 0)
            v += a3 * sin(p3 * k);
        if (p4 > 0)
            v += a4 * sin(p4 * k);
    } else if (ifd < 0 && stim_type != 4) {
	v = (mts == 0) ? v0 : 0;	/* impulse stimulus */
    } else {
	k = mts * ts;
	if (k >= sts) {			/* beyond EOF */
	    v = 0;
	} else if (n_up > 0) {
	    v = interp(k);		/* LPF interpolate (w/ wrap) */
	} else {
	    v = fetch(k);
	}
    }
    g = pa_.stgain;
    if (pa_.ntsgat > 0) {
	t = (mts + pa_.icyc * pa_.ntsc) / pa_.ntsgat;
	if (t > 0 && t < 1)
	    g *= (1 - cos(pi * t)) / 2;
    }
    return (v * g);
}

static double
fetch_ils(long k)
{
    int     kpnt, kblk, ival;

    kblk = k / BS + 1;
    kpnt = k % BS;
    if (kblk == iblk) {
	ival = ibuf[kpnt];
    } else if (kblk == jblk) {
	ival = jbuf[kpnt];
    } else if (iblk < jblk) {
	iblk = kblk;
	(void) rdblk(ifd, iblk, ibuf, 1);
	ival = ibuf[kpnt];
    } else {
	jblk = kblk;
	(void) rdblk(ifd, jblk, jbuf, 1);
	ival = jbuf[kpnt];
    }

    return ((double)ival);
}

static double
fetch_stm(long k)
{
    double v;

    lseek(ifd, stm_doff + stm_dsiz * k, 0);
    rd_stm_val(ifd, stm_dtyp, &v);
    return (v);
}

static double
fetch_imp(long k)
{
    double v = 0;
    static short im30[64] = {
             0,     1,     1,   278,  4074, 18370, 32000, 16341,-11939, -8468,
          8759,  2506, -6497,  1069,  3817, -2477, -1400,  2357,  -195, -1517,
           884,   604,  -910,    36,   609,  -329,  -256,   355,     0,  -245,
           122,   108,  -138,    -5,    98,   -45,   -45,    54,     4,   -39,
            16,    19,   -21,    -2,    15,    -6,    -7,     8,     1,    -6,
             2,     3,    -3,     0,     2,     0,    -1,     1,     0,    -1,
             0,     0,     0,     0
    };

    if (k < 64)
        v = im30[k];
    return (v);
}

double
fetch(long k)
{
   double v = 0;

    while (k < 0)
	k += nssc;
    if (k > nssc)
	k %= nssc;
    if (k > nsmp)
	v = 0;
    else if (stim_type == 1)
	v = fetch_ils(k);
    else if (stim_type == 2)
	v = fetch_stm(k);
    else if (stim_type == 3)
	v = fetch_stm(k);
    else if (stim_type == 4)
	v = fetch_imp(k);
    return (v);
}

/*********************************************************************
 *		blkio routines
 *********************************************************************/

int
opnblk(char *file)
/* open a file in raw mode for block io and return a file descriptor. */
{
    short   fd, oflg;

    oflg = O_RDONLY | O_BINARY;
    fd = open(file, oflg, 0);
    return (fd);
}

int
rdblk(short fd, short block, short *ibuf, short blocks)
/* rdblk - read from fd, blocks into ibuf starting at block */
{
    long    offset;
    unsigned int count, rd, i;

    offset = block * 512L;
    (void) lseek(fd, offset, 0);/* seek from block 0 */
    count = 512 * (unsigned int) blocks;
    rd = read(fd, (char *) ibuf, count);
    if (swap_bytes)
	swab((char *) ibuf, (char *) ibuf, rd);
    if (rd < count)
	for (i = (rd < 0) ? 0 : rd / 2; i < (count / 2); i++)
	    ibuf[i] = 0;

    return (rd);
}

void
clsblk(short fd)
{
    close(fd);
}

/*********************************************************************
 *		interp routines routines
 *********************************************************************/

void
compute_lpf()
{
    int     nn, i;
    float   a, w, p;
    static float s = 1;
    static float pi = 3.1415927;

    nn = n_up * n_fw + 1;
    if (hh == NULL) {
	if ((hh = (float *) malloc(nn * sizeof(float))) == NULL) {
	    perror("mem alloc for h[] failed");
	    exit(1);
	}
    }
    hh[0] = s;
    for (i = 1; i < nn; i++) {
	p = (float) i *pi / (nn - 1);
	w = 0.42 + 0.5 * cos(p) + 0.08 * cos(2 * p);	/* Blackman window  */
	a = (float) i *pi / n_up;
	hh[i] = s * w * sin(a) / a;	/* lpf impulse response */
    }

}

double
interp(kk)
long int kk;
{
    int     i, k2, nn;
    long    k, k1;
    double  sum;

    k1 = kk / n_up;
    k2 = (int) (kk % n_up);
    sum = 0;

    nn = n_fw;
    k = k1;
    for (i = k2; nn-- > 0; i += n_up)
	sum += hh[i] * fetch(k--);

    nn = n_fw;
    k = k1 + 1;
    for (i = n_up - k2; nn-- > 0; i += n_up)
	sum += hh[i] * fetch(k++);

    return (sum);
}
