/* tdmstp.c - time-step subroutines for time-domain model */

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <time.h>
#include <math.h>
#include "tdmdef.h" 
#include "cdata.h"
#include "cimped.h"
#include "cmatrx.h"
#include "cmider.h"
#include "cparam.h"
#include "cstate.h"
#include "ctdmio.h"

#define limit(a,b,c)	(((a)>(b))?(a):((c)<(b))?(c):(b))

struct _grad {
    float   ddd_[NSTVmax];
} cur, nxt; 
struct _state nx_;

double  stimul();
void    curstp();
void    display();
void    gbe();
void    nxtstp();
void    savest();
void    stmcls();
void    stmopn();
void    tdmopn();
void    tdmcls();
void    wrdata();
void    wrpar();

static void accel();
static void update();
static void endstp();
static void inctim();

static long isav = 300;		/* save state every 5 minutes */
static long tsav = 0;

/* timstp - compute state for each timestep */

void
timstp()
{
    if (pa_.icyc >= pa_.ncyc)
	return;
    if (pa_.ntsw <= 0)		/* check whether to write data */
	wridat = 0;
    tdmopn();			/* open tdm data file */
    while (!brkflg) {		/* time step loop */
	curstp();		/* compute pressure & acceleration */
	nxtstp();		/* advance to next time step */
    }
    tdmcls();			/* close tdm data file */
}

void
tdmopn()
{
    static void onintr();

    signal(SIGINT, onintr);	/* catch interrupts */
    signal(SIGHUP, onintr);	/* catch hangups */
    sfp = fopen(sfn, "r+b");
    dfp = fopen(dfn, "r+b");
    stmopn();
    if (pa_.ifrmt == 0)
	wrpar(lfp);		/* write parameters to data file */
    savest();			/* save initial model state */
    tsav = time(0L) + isav;
}

void
curstp()
{
    double  stim;

    stim = stimul(pa_.itsc);
    accel(&st_, &cur, stim);	/* compute acceleration */
    if (pa_.itsw == 0 && wridat && pa_.icyc < pa_.ncyc)
	wrdata();
    display();
    if (dbgflg[0]) {
	fprintf(lfp, "  itsc,icyc,itsw,itsf = %8ld%8ld%8ld%8ld\n",
	    pa_.itsc, pa_.icyc, pa_.itsw, pa_.itsf);
    }
    if (dbgflg[1]) {
	fprintf(lfp, "  stim,vep,ped,pst = %13.4g%13.4g%13.4g%13.4g\n",
	    stim, vep, ped, pst);
    }
    if (dbgflg[2]) {
	fprintf(lfp, "  dst,ddst,dddst = %10.2g%10.2g%10.2g\n",
	    st_.d_[me_st], st_.dd_[me_st], cur.ddd_[me_st]);
	fprintf(lfp, "  dma,ddma,dddma = %10.2g%10.2g%10.2g\n",
	    st_.d_[me_ma], st_.dd_[me_ma], cur.ddd_[me_ma]);
    }
    if (time(0L) > tsav) {
	savest();		/* save current model state */
	tsav += isav;
    }
}

void
nxtstp()
{
    int     i;
    double  stim;

    update(0);			/* update next state vectors */
    stim = stimul(pa_.itsc + 1);
    for (i = 0; i < pa_.nimp; i++) {
	accel(&nx_, &nxt, stim);/* compute acceleration */
	update(1);		/* update next state vectors */
    }
    endstp();			/* end of time-step */
    inctim();			/* increment to next time-step */
}

void
tdmcls()
{
    savest();			/* save final model state */
    stmcls();
    fclose(sfp);
    fclose(dfp);
}

static void
onintr()
{
    signal(SIGINT, SIG_IGN);	/* ignore interrupts */
    fprintf(stderr, "interrupt!\n");
    brkflg++;
}

/* pressr1d - compute 1D fluid pressure */

static double
pressr1d(double stim, float smest, float *q)
{
    int i, k, nn = pa_.n - 1;

    /* set up force function q */
    pp[0] = (abmom[0] * q[0] - alfx * smest) * abas + stim * pbas;
    for (i = 1; i < nn; i++)
        pp[i] = abmom[i] * q[i];
    pp[nn] = abmom[nn] * q[nn] * aapx;

    /* solve for pressure ' a * p = q ' */
    pp[0] = aa2[0] * pp[0] * abas + pp[0] * pbas;
    for (k = 1; k < nn; k++)
        pp[k] = aa2[k] * (pp[k] - aa1[k] * pp[k - 1]);
    pp[nn] = aa2[nn] * (pp[nn] - aa1[nn] * pp[nn - 1]) * aapx + q[nn] * papx;
    for (k = nn - 1; k > 0; k--)
        pp[k] -= aa3[k] * pp[k + 1];
     pp[0] -= aa3[0] * pp[1] * abas;

    return (pp[0]);
}

/* pressr2d - compute 2D fluid pressure */

static double
pressr2d(double stim, float smest, float *q)
{
    float qst, tk[Mmax];
    int i, k, m, mm1, n;

    n = pa_.n;
    m = pa_.m;
    mm1 = m - 1;

    /* zero pressure array */
    for (k = 0; k < n; k++)
        for (i = 0; i < m; i++)
            pp[i + k * m] = 0;

    /* set up force function q */
    qst = stim * pbas - alfx * smest * abas;
    pp[0] = qst + abmom[0] * q[0];
    for (i = 1; i < m; i++)
        pp[i] = qst;
    for (k = 1; k < n; k++)
        pp[k * m] = abmom[k] * q[k];

    /* solve for pressure ' a * p = q ' */
    gbe(aaa, pp, tk, m, n, pa_.basacc, pa_.apxacc);

    qst = (pp[0] + pp[mm1]) / 2;
    for (i = 1; i < mm1; i++)
	qst += pp[i];
    qst /= mm1;
    for (k = 0; k < n; k++)
        pp[k] = pp[k * m];

    return (qst);
}

/* accel - compute pressure and acceleration */

static void
accel(struct _state *st, struct _grad *gr, double stim)
{
    float   tvgam, doref, dosat, dotmp, ddotmp, s3tmp, s4tmp, s5tmp, r1tmp;
    float   do1, do2, do3, do4, oc1, oc2, ka;
    float  *dme, *ddme, *dddme;
    float  *d1, *dd1, *ddd1;
    float  *d2, *dd2, *ddd2;
    float  *vm, *dvm;
    float  *s1, *s2;
    float  sme[NME], sco, pco, prw;
    int     i, j, nn;

    dme = st->d_ + ST_OFF_ME;
    d1 = st->d_ + ST_OFF_1;
    d2 = st->d_ + ST_OFF_2;
    ddme = st->dd_ + ST_OFF_ME;
    dd1 = st->dd_ + ST_OFF_1;
    dd2 = st->dd_ + ST_OFF_2;
    dddme = gr->ddd_ + ST_OFF_ME;
    ddd1 = gr->ddd_ + ST_OFF_1;
    ddd2 = gr->ddd_ + ST_OFF_2;
    s1 = ddd1;
    s2 = ddd2;
    dvm = st->dvohc;
    vm = st->vohc;

    nn = pa_.n - 1;
    oc1 = 1 / pa_.ohccmp;
    oc2 = pa_.ohccmp;

    /*.................................................. state force */

    /* middle-ear and coupler */

    for (i = 0; i < NME; i++) {
	sme[i] = -stim * smes[i];
        for (j = 0; j < NME; j++) {
	    sme[i] += smek[i][j] * dme[j] + smer[i][j] * ddme[j];
	}
    } 

    /* cochlea */

    if (nn <= 0) {
	s2[0] = pa_.rco * (ddme[me_st] - dd2[0]) * ame[me_st]
	    - pa_.rve * dd2[0] * ame[me_st];
    } else if (pa_.mmeq == -1) {		/* MOH83 */
	for (i = 0; i <= nn; i++) {
	    s1[i] = -(z1k[i] * d1[i] + (z1r[i] + g1[i]) * dd1[i])
	    	+ (z3k[i] * d2[i] + pa_.gam * z3r[i] * dd2[i]);
	    s2[i] = -(z2k[i] * d2[i] + (z2r[i] + g2[i]) * dd2[i])
	    	+ (z3k[i] * d1[i] + pa_.gam * z3r[i] * dd1[i]);
	}
    } else if (pa_.mmeq == 0) {			/* JASA81 */
	for (i = 0; i < nn; i++) {
	    r1tmp = z1r[i];
	    if (nltype)                 /* nonlinear dampling */
	        r1tmp *= 1 + (g1[i] + g2[i] * dd1[i]) * dd1[i];
	    s1[i] = -(z1k[i] * d1[i] + r1tmp * dd1[i]);
	}
	s1[nn] = -(z1k[nn] * d1[nn] + z1r[nn] * dd1[nn]);
    } else if (pa_.mmeq == 1) {			/* MOH90 */
	for (i = 0; i < nn; i++) {
	    dotmp = d1[i] - d2[i];
	    ddotmp = dd1[i] - dd2[i];
            tvgam = pa_.gam;
	    if (nltype == 1) {	/* hard limit */
		dosat = limit(-pa_.d3max, dotmp, pa_.d3max);
		if (dotmp != dosat)
                    tvgam *= dosat / dotmp;
	    }
	    s3tmp = (z3k[i] * dotmp + z3r[i] * ddotmp);
	    s4tmp = (z4k[i] * dotmp + z4r[i] * ddotmp) * tvgam;
	    s1[i] = -(z1k[i] * d1[i] + z1r[i] * dd1[i]) - s3tmp + s4tmp;
	    s2[i] = -(z2k[i] * d2[i] + z2r[i] * dd2[i]) + s3tmp;
	}
	s1[nn] = -(z1k[nn] * d1[nn] + z1r[nn] * dd1[nn]);
	s2[nn] = 0;
    } else if (pa_.mmeq == 2) {			/* MOH93 */
	doref = pa_.d3max;
	for (i = 0; i < nn; i++) {
	    dotmp = d1[i] - d2[i];
	    ddotmp = dd1[i] - dd2[i];
            tvgam = pa_.gam;
	    if (nltype == 1) {	/* soft limit */
		if (dotmp < 0) {
		    tvgam *= doref / (doref - dotmp);
		} else {
		    tvgam *= doref / (doref + dotmp);
		}
	    }
	    dosat = dotmp * tvgam;
	    dvm[i] = (dosat * g1[i] - vm[i]) / t1[i];
	    s3tmp = (z3k[i] * dotmp + z3r[i] * ddotmp);
	    s4tmp = (z4k[i] * g2[i] * vm[i]);
	    s1[i] = -(z1k[i] * d1[i] + z1r[i] * dd1[i]) - s3tmp;
	    s2[i] = -(z2k[i] * d2[i] + z2r[i] * dd2[i]) + s3tmp + s4tmp;
	}
	s1[nn] = -(z1k[nn] * d1[nn] + z1r[nn] * dd1[nn]);
	s2[nn] = 0;
	dvm[nn] = 0;
    } else if (pa_.mmeq == 3) {		/* MOH96 */
	doref = pa_.d3max;
	for (i = 0; i < nn; i++) {
	    dotmp = d1[i] - d2[i];
	    ddotmp = dd1[i] - dd2[i];
	    do1 = dotmp / doref;
	    do2 = do1 * do1;
	    do4 = do2 * do2;
	    tvgam = pa_.gam * pow(1 + do2 / 2 + do4, -0.25);
	    dosat = dotmp * tvgam;
	    dvm[i] = (dosat * g1[i] - vm[i]) / t1[i];
	    s3tmp = (z3k[i] * dotmp + z3r[i] * ddotmp);
	    s4tmp = (z4k[i] * g2[i] * vm[i]);
	    s1[i] = -(z1k[i] * d1[i] + z1r[i] * dd1[i]) - s3tmp;
	    s2[i] = -(z2k[i] * d2[i] + z2r[i] * dd2[i]) + s3tmp + s4tmp;
	}
	s1[nn] = -(z1k[nn] * d1[nn] + z1r[nn] * dd1[nn]);
	s2[nn] = 0;
	dvm[nn] = 0;
    } else if (pa_.mmeq == 4) {			/* MOH99 */
	doref = pa_.d3max;
	for (i = 0; i < nn; i++) {
	    dotmp = d1[i] - d2[i];
	    ddotmp = dd1[i] - dd2[i];
	    do1 = dotmp / doref;
	    do4 = pow(fabs(do1), oc1);
	    tvgam = pa_.gam / pow(1 + do4, oc2);
	    dosat = dotmp * tvgam;
	    dvm[i] = (dosat * g1[i] - vm[i]) / t1[i];
	    s3tmp = (z3k[i] * dotmp + z3r[i] * ddotmp);
	    s4tmp = (g2[i] * vm[i]);
	    s1[i] = -(z1k[i] * d1[i] + z1r[i] * dd1[i]) - s3tmp;
	    s2[i] = -(z2k[i] * d2[i] + z2r[i] * dd2[i]) + s3tmp + s4tmp;
	}
	s1[nn] = -(z1k[nn] * d1[nn] + z1r[nn] * dd1[nn]);
	s2[nn] = 0;
	dvm[nn] = 0;
    } else if (pa_.mmeq == 5) {			/* MOH99+ */
	doref = pa_.d3max;
	for (i = 0; i < nn; i++) {
	    dotmp = d1[i] - d2[i];
	    ddotmp = dd1[i] - dd2[i];
	    do1 = fabs(dotmp / doref);
	    do2 = do1 * do1;
	    do3 = do1 * do2;
	    do4 = do2 * do2;
	    tvgam = pa_.gam * pow(1 + do1 + do2/2 + do3/6 + do4/24, -0.25);
	    dosat = dotmp * tvgam;
	    dvm[i] = (dosat * g1[i] - vm[i]) / t1[i];
	    s3tmp = (z3k[i] * dotmp + z3r[i] * ddotmp);
	    s4tmp = (g2[i] * vm[i]);
	    s1[i] = -(z1k[i] * d1[i] + z1r[i] * dd1[i]) - s3tmp;
	    s2[i] = -(z2k[i] * d2[i] + z2r[i] * dd2[i]) + s3tmp + s4tmp;
	}
	s1[nn] = -(z1k[nn] * d1[nn] + z1r[nn] * dd1[nn]);
	s2[nn] = 0;
	dvm[nn] = 0;
    } else if (pa_.mmeq == 6) {			/* MOH02- */
	for (i = 0; i <= nn; i++) {
	    if (pa_.gam == 0 || pa_.dhbmax == 0 || d1[i] == 0) {
	        tvgam = pa_.gam;
	    } else {
	        dotmp = 1 / (1 + exp(- d1[i] / pa_.dhbmax));
	        tvgam = pa_.gam * 4 * pa_.dhbmax * (dotmp - 0.5) / d1[i];
	    }
	    ka = z3k[i] / (1 + z4k[i]);
	    s1[i] = -((z1k[i] + z2k[i]) * d1[i] + (z1r[i] + g1[i]) * dd1[i])
	    	- (z2k[i] * d2[i]);
	    s2[i] = -(z2k[i] * d2[i] + (z2r[i] + g2[i]) * dd2[i])
	    	- ((z2k[i] - tvgam * ka) * d1[i]);
	}
    } else if (pa_.mmeq == 7) {			/* MOH02 */
	#define BF1(x) (1/(1+exp(-bfa1*x)))
	#define BF2(x) (1/(1+exp(bfa2*(bfx2-x))*(1+exp(bfa1*(bfx1-x)))))
	double bfzer, bfslp;
	double bfa1, bfa2, bfx1, bfx2;
	static double pm = 1e-10;	/* 1 picometer */ 
	static double gta1 = 0.065;
	static double gta2 = 0.016;
	static double gtx1 = 24;
	static double gtx2 = 41;
	static double gtx0 = 23.982;
	static double gtxs = 24.5;
	if (pa_.dhbmin == 0 && pa_.dhbmax != 0) {
	    bfa1 = 1 / pa_.dhbmax;
	    bfzer = BF1(0);
	    bfslp = (BF1(pm) - BF1(-pm)) / (2 * pm);
	} else if (pa_.dhbmin > 0) {
	    gtxs = pa_.dhbmax;
	    bfa1 = gta1 / pa_.dhbmin;
	    bfa2 = gta2 / pa_.dhbmin;
	    bfx1 = (gtx1 - gtx0 - gtxs) * pa_.dhbmin;
	    bfx2 = (gtx2 - gtx0 - gtxs) * pa_.dhbmin;
	    bfzer = BF2(0);
	    bfslp = (BF2(pm) - BF2(-pm)) / (2 * pm);
	}
	for (i = 0; i <= nn; i++) {
	    dotmp = d1[i] + d2[i];
	    ddotmp = dd1[i] + dd2[i];
	    s3tmp = z3k[i] * dotmp + z3r[i] * ddotmp;
	    if (pa_.gam == 0 || d1[i] == 0) {
		dosat = 0;
	    } else if (pa_.dhbmax < 0) {
#ifdef comp4
		do1 = d1[i] / pa_.dhbmax;
		do2 = do1 * do1;
		do4 = do2 * do2;
		dosat = pa_.gam * d1[i] * pow(1 + do2 / 2 + do4, -0.25);
#else
		do1 = pa_.dhbmin;
		do2 = pow(fabs(d1[i] / pa_.dhbmax), do1);
		dosat = pa_.gam * d1[i] * pow(1 + do2, -1/do1);
#endif
	    } else if (pa_.dhbmin != 0) {
		dosat = pa_.gam * (BF2(d1[i]) - bfzer) / bfslp;
	    } else if (pa_.dhbmax != 0) {
		dosat = pa_.gam * (BF1(d1[i]) - bfzer) / bfslp;
	    } else {
	        dosat = pa_.gam * d1[i];
	    }
	    if (t1[i] == 0) {
	    	s4tmp = z4k[i] * dosat;
	    } else {
		dvm[i] = (z4k[i] * dosat - (1 + g1[i]) * vm[i]) / t1[i];
		s4tmp = vm[i];
	    }
	    s5tmp = 0;
	    if (i > 0)
		s5tmp += z5k[i] * (d1[i] -  d1[i - 1]);
	    if (i < nn)
		s5tmp += z5k[i] * (d1[i] -  d1[i + 1]);
	    s1[i] = -(z1k[i] * d1[i] + z1r[i] * dd1[i] + s3tmp + s5tmp);
	    s2[i] = -(z2k[i] * d2[i] + z2r[i] * dd2[i] + s3tmp - s4tmp);
	}
    } else {				/* default */
	fprintf(lfp, "Micromechanics not implemented for mmeq=%d.\n", pa_.mmeq);
	exit(1);
    }

    /*.................................................. pressure */

    /* cochlea */

    prw = pa_.rrw * ddme[me_st] * ame[me_st]
        + pa_.krw * dme[me_st] * ame[me_st];
    sco = sme[me_st] + prw;
    if (nn <= 0) {
        pco = pp[0] = pa_.rco * (ddme[me_st] - dd2[0]) * ame[me_st] * abas
	    + stim * pbas;
    } else if (pa_.m == 1) {
        pco = pressr1d(stim, sco, s1);
    } else {
        pco = pressr2d(stim, sco, s1);
    }
    pst = pco + prw;

    /*.................................................. acceleration */

    /* middle-ear & coupler */

    sme[me_st] += pst;
    for (i = 0; i < NME; i++) {
        dddme[i] = 0;
        for (j = 0; j < NME; j++) {
            dddme[i] -= smem[i][j] * sme[j];
        }
    }
    vep = smes[me_ep] ? stim : 0; 
    ped = peds * stim;
    for (i = 0; i < NME; i++) {
        ped += pedm[i] * dddme[i] + pedr[i] * ddme[i] + pedk[i] * dme[i];
    }

    /* cochlea */

    for (i = 0; i <= nn; i++)
        ddd1[i] = (s1[i] - pp[i]) / z1m[i];
    if (nn <= 0 || pa_.mmeq != 0)
        for (i = 0; i <= nn; i++)
            ddd2[i] = s2[i] / z2m[i];

}

/* update - update next state */

static void
update(f)
{
    float   dto2, dtt;
    int     i;

    dtt = pa_.dt;
    dto2 = pa_.dt / 2;
    if (f == 0) {			/* modified Sielecki */
	for (i = 0; i < NSTV; i++) {
	    nx_.dd_[i] = st_.dd_[i] + cur.ddd_[i] * dtt;
	    nx_.d_[i] = st_.d_[i] + (st_.dd_[i] + nx_.dd_[i]) * dto2;
	}
	for (i = 0; i < pa_.n; i++) {
	    nx_.vohc[i] = st_.vohc[i] + st_.dvohc[i] * dtt;
	}
    } else {				/* modified Heun */
	for (i = 0; i < NSTV; i++) {
	    nx_.dd_[i] = st_.dd_[i] + (cur.ddd_[i] + nxt.ddd_[i]) * dto2;
	    nx_.d_[i] = st_.d_[i] + (st_.dd_[i] + nx_.dd_[i]) * dto2;
	}
	for (i = 0; i < pa_.n; i++) {
	    nx_.vohc[i] = st_.vohc[i] + (st_.dvohc[i] + nx_.dvohc[i]) * dto2;
	}
    }
}


static void
endstp()
{
    float   wncc, s, k;
    float  *d1, *d2;
    int     i;
    static float rc = 5e-4;	/* IHC time constant */
    static float e = 1e-8;	/* IHC transduction breakpoint (cm) */
    //static float nrgn = 2.5e6;  /* neural rate gain */

    /* middle-ear and cochlea */
    for (i = 0; i < NSTV; i++) {
	st_.dd_[i] = nx_.dd_[i];
	st_.d_[i] = nx_.d_[i];
    }

    if (st_.pedmax < -ped)
        st_.pedmax = -ped;
    if (st_.pedmax < ped)
        st_.pedmax = ped;
    st_.pedsum += ped;
    st_.pedssq += ped * ped;
    st_.pednum += 1;

    /* outer hair cell */
    for (i = 0; i < pa_.n; i++) {
	st_.vohc[i] = nx_.vohc[i];
	st_.dvohc[i] = nx_.dvohc[i];
    }

    /* inner hair cell - leaky integrator model using first-order
     * difference equation from Gold and Rader (1969) p 21 */
    d1 = st_.d_ + ST_OFF_1;
    d2 = st_.d_ + ST_OFF_2;
    k = exp(-pa_.dt / rc);
    wncc = (pa_.n == 1) ? 1 : 0;
    for (i = 0; i < pa_.n; i++) {
	if (pa_.ihceq == 1) {			/* MOH02 */
	    s = d1[i];
	    s = (s < 0) ? 0 : s;		/* rectify */
	    s *= s * s / ((e + s) * (e + s));	/* cube threshold */
	} else {				/* MOH99 */
	    s = d1[i] - d2[i];
	    s = (s < 0) ? 0 : s;		/* rectify */
	    s *= s * s / ((e + s) * (e + s));	/* cube threshold */
	}
	st_.cc[i] = k * st_.cc[i] + s;
	st_.sc[i] = st_.cc[i];
	wncc += st_.sc[i];
    }
    st_.wnap = pa_.nrgn * wncc;
}


static void
inctim()
{
    pa_.itsc++;
    pa_.itsw++;
    if (pa_.itsc >= pa_.ntsc) {
	pa_.itsc = 0;
	pa_.icyc++;
    }
    if (pa_.icyc >= pa_.ncyc) {
	brkflg++;
    }
    if (pa_.itsw >= pa_.ntsw) {
	pa_.itsw = 0;
	pa_.itsf++;
    }
    if (pa_.itsf >= pa_.ntsf)
	pa_.itsf = 0;
}
